''=============================================================================
'' file:     i2c_123.spin    version 1.0
'' description:   Propeller I spin methods to set up an i2c buss that can be shared by various devices/objects.
'' author:   Thomas Tracy Allen, EME Systems LLC
'' Copyright (c) 2013 EME Systems LLC
'' See end of file for terms of use.
'' This was inspired and informed by several other i2c objects on the forums and in the OBEX, now tuned for my own way of working.
'' Thank you.
''
'' Inits an i2c bus so that different devices can access it at their invidual buss addresses but sharing the pins.
'' Objects variables are held in DAT space so that different objects can declare it as a single instance.
'' Only one object (usually a top object) will call the Init method, other objects declare it OBJ but do not init. (no harm to re Init tho)

'' Both sda and scl need pullup resistors. When in idle state the pins are released to the pullup state.
'' If other i2c objects use the same buss pins, they should float both the sda and scl lines when not active.
'' Parameter "devID" is the 7-bit address of the device on the i2c buss.
'' NOTE NOTE NOTE: be sure to use the 7-bit address, not shifted to 8 bits to make room for r/w bit.
'' The driver will provide the lsb 0=write, 1=read
''
'' This object automatically supports clock stretching for data reads (scl is tested before reading each elementary byte)
''
'' parameter "devReg" is a device command, pointer or address register.
'' For one byte device, use positive value.
''      i2c.CommandAddress(devID,CONFIG)   ' sets the 1 byte config register
'' For two-byte word address (e.g. memory), make it a negative value, which will be written as a positive value, msb first
''      i2c.CommandAddress(devID,-address)  ' sets the two byte starting address for a memory chip
'' Larger memories (>64 kbytes) require address bits in the device ID, There are various schemes, which should be handled by the memory routines.
''
'' This object supports clock stretching by the master, and this occurs automatically.
'' The potential hold occurs between the ACK of one byte and the first clock low to high transition of the next byte.
'' There is a timeout parameter, which defaults to 100ms, with a PUB method to change that to suit the slowest device.
'' The object does not implement polling, too many possibilities.  Any polling routines should also implement a timeout to forestall buss lockups.
''
'' String and numeric methods that send or receive data take an an additional parameter for number of bytes to transfer.
'' If positive, the bytes are transferred least significant byte first to/from Propeller memory.
''      i2c.WriteString(devId, devReg, dataPtr, byteCount)
'' If negative, the bytes are transferred in reverse order, for bigEndian/smallEndian matchup.
''     result := i2c.ReadNumber(%100_0000, 0, -2)      ' ReadNumber(devID, devReg, devMode, numSize)
'' Device ID=%1000_000, command=0 one byte, read data two bytes msb first but return result small-endian to propeller.

'' Some methods return a count of the total number of ackbits detected during the transaction.  Can be tested against the number expected.
''
'' This object also supports locking the buss. This feature has not been tested extensively.
'' SetLock should be invoked before start of a transaction, ClearLock after.  There is no timeout on the lock.
'' In order for the buss lock to be effective. all i2c traffic for this buss should be routed via this one object.
''
'' Multiple instances of this object can be declared to open i2c busses on different pins.
'' If multiple instances of this object are required, change the DAT variable "instance" to a different value.

 '' Tracy Allen, EME Systems

''  tested and written into drivers for: ADS1115, TMP102, ISL12020, PCA9554, PCA9536, SHT25 (uses clock stretching), MAX1236 ADC.

CON
   ACK      = 0                        ' I2C Acknowledge
   NAK      = 1                        ' I2C No Acknowledge

   PULLUPHIGH = 0                      ' for dira, pin is open and pulled high by resistor
   DRIVELOW = 1                       ' for dira, output to drive pin low
   LOW = 0                             ' for outa, will be left low, all action is via dira

VAR
  ' see DAT section.  This object can be declared within several other objects to utilize the same i2c bus.
  ' in order to create a separate i2c bus with its own pins, change the DAT variable instance to a different number.

CON  ' low level routines

PUB Init(_scl, _sda)            ' what pins?
  sda := _sda
  scl := ||_scl                 ' negative pin value flags desire not to twiddle
  dira[scl] := PULLUPHIGH       ' both scl and sda are pulled high by resistor
  dira[sda] := PULLUPHIGH
  outa[scl] := LOW              ' leave a zero in the scl and sda output latches.
  outa[sda] := LOW              ' so turning pin to output will drive the pin low.
  if _SCL>-1                    ' entering with _SCL negative causes this to skip the Twiddle (some devices don't like to be twiddled).
    Twiddle
  return ina[sda..scl]          ' should return 3, otherwise buss is locked.

PUB Twiddle
   repeat 9
      if ina[sda]               ' want to clear the bus
         quit                          '
      dira[scl] := DRIVELOW     ' Put out up to 9 clock pulses
      dira[scl] := PULLUPHIGH   ' leave scl pulled high
   result := ina[sda..scl]      ' correct return value is 3, otherwise i2c buss is locked.

PUB Start                       ' sda goes HIGH to LOW with scl HIGH
   dira[scl] := PULLUPHIGH
   dira[sda] := DRIVELOW
   dira[scl] := DRIVELOW
   dira[sda] := PULLUPHIGH      ' allow sda to track

PUB Stop                                ' sda goes LOW to HIGH with scl High,
   dira[sda] := DRIVELOW                              ' normally don't need this.
   dira[scl] := PULLUPHIGH                           ' scl pulled high
   dira[sda] := PULLUPHIGH                           ' sda pulled high

PUB Write(bite) : ackbit | ticks
'' Write i2c byte up through detection of ack/nak
'' Bite is output MSB first.  Sda changes only when clk is low.
'' scl and sda are both left in the open state, pulled high.
   ackbit := 0
   bite <<= 24
   repeat 8                     ' Output data to sda
     dira[scl] := DRIVELOW                        ' scl driven low
     dira[sda] := !(bite <-= 1) ' can change dta now with scl low. Note inversion, data=1 is dir=0. totating msb 1st into bit 0.
     dira[scl] := PULLUPHIGH    ' release clk high
   dira[scl] := DRIVELOW        ' now for ack/nak bit, clk low
   dira[sda] := PULLUPHIGH      ' and sda input to allow ack/nak from device
   dira[scl] := PULLUPHIGH      ' clock high to capture ack/nak
   ackbit := ina[sda]           ' read ack/nak
  dira[scl] := DRIVELOW         ' scl driven low, will be followed by next byte, or stop.


PUB Read(ackbit) | ticks
  dira[sda] := PULLUPHIGH     ' normally don't need this.
  dira[scl] := PULLUPHIGH       ' scl back high
  if ina[scl] == 0              ' allow for clock stretching
    ticks := cnt
    repeat while (ina[scl]==0) and (cnt - ticks  < timeout)
  repeat 8                      ' Receive data from sda
      dira[scl] := PULLUPHIGH   ' scl back high
      result := (result << 1) | ina[sda]
      dira[scl] := DRIVELOW     ' scl low, allow slave to set data
  dira[sda] := !ackbit          ' Output ACK/NAK to sda
  dira[scl] := PULLUPHIGH
  dira[scl] := DRIVELOW         ' scl low, allow local to change data, or ready for stop bit
  ' returns result, a byte


PUB SetTimeout(ms)
' This is the timeout for clock stretching
  timeout := clkfreq/1000 * ms

CON  ' device dependent routines.  Open a device by sending its device ID (busAddress), and command/pointer/address value.
' The command/pointer/address value is 1 byte by default.  For 2 byte address (memory), msb 1st, use negative value for devReg.
' OpenDevice(..) leaves the i2c bus open for continuation write or reads.  CommandAddress(..) is completed with Stop.
PUB OpenDevice(devID, devReg) : ackbits   ' this sends the devID and a pointer, stays open in preparation for further action.
  Start
  ackbits := 1 - Write(devID<<1)
  if devReg < 0
    ||devReg
    ackbits += 1 - Write(devReg.byte[1])                ' devMode contains the number of address bytes needed
  ackbits += 1 - Write(devReg.byte)


'SetPointerRegister            ' DEPRECATE this sets a device command, pointer register or starting address.
'SetDeviceRegister
PUB CommandAddress(devID, devReg) : ackbits
' this is the same as open device as a standalone transaction.  Loads the one or two byte command or starting address register.
' Use negative devReg value for two bytes.
  ackbits := OpenDevice(devID, devReg)
  Stop


CON
' Writes or reads a string having a certain length (byteCount)
' Use negative byteCount to write or read backwards, e.g., small-endian to large-endian or vice versa.
' Also useful for things like reversing values of time/date registers in real time clock.
' devReg is a command, pointer or address register in the device.  Use a positive value for 1 byte, negative value for 2 bytes.
' ReReadString is a continuation possible in some devices, where the address pointer register autoincrements or remains constant.
' Both the write and read methods return the number of ackbits detected, which can then be checked against the number expected.
PUB WriteString(devId, devReg, dataPtr, byteCount) : ackbits
' if byteCount is negative, write successive data bytes backwards (e.g. to send msb first into small-endian variables)
  ackbits := OpenDevice(devID, devReg)
  if byteCount<0
    || byteCount
    dataptr += byteCount
    repeat byteCount
      ackbits += 1 - Write(byte[--dataPtr])
  else
    repeat byteCount
      ackbits += 1 - Write(byte[dataPtr++])      '' write successive bytes small endian like prop.
  Stop   ' ackbits==byteCount + devRegbytes + 1

PUB ReadString(devID, devReg, dataPtr, byteCount) : ackbits
  ackbits := OpenDevice(devID, devReg)                      ' could bail out if this does not equal regcount+1
  ackbits += ReReadString(devID, dataPtr, byteCount)
  ' this returns the total number of acks, and the string is written to the dataPtr address

PUB ReReadString(devID, dataPtr, byteCount) : ackbits | ticks
  Start
  ackbits := 1 - Write(devID<<1 | 1)
  if byteCount < 0
    || byteCount
    dataPtr += byteCount - 1
    repeat byteCount - 1
      byte[dataPtr--] := Read(ACK)
  else
    repeat byteCount - 1
      byte[dataPtr++] := Read(ACK)
  byte[dataPtr] := Read(NAK)
  Stop

CON
' the number in in the following methods can be from 1 to 4 bytes, numsize should match.
' Use negative numSize value to write or read backwards, e.g., small-endian to large-endian or vice versa.
' devMode is the number of bytes in the devReg address registor, and it can be 0, 1 or 2 bytes.
' ReRead is a continuation possible in some devices, where the address pointer register autoincrements or remains constant.
PUB WriteNumber(devID, devReg, number,numSize) : ackbits
  ackbits += WriteString(devID, devReg, @number,numSize)

PUB ReadNumber(devID, devReg, numSize) ' datacount can be 1, 2, 3 or 4 (to fit in the long, result)
  ReadString(devID, devReg, @result, numSize)

PUB ReReadNumber(devID, numSize)
  ' use this when the pointer or command register is already set or autoincrements.
  ReReadString(devID, @result, numSize)

' the above Number methods can be used for bytes (numSize=1), but the following methods are quicker.
PUB WriteByte(devID, devReg, bite) : ackbits
  Start
  ackbits := 1 - Write(devID<<1)
  ackbits += 1 - Write(devReg)
  ackbits += 1 - Write(bite)
  Stop          ' should return ackbits==3

PUB ReadByte(devID, devReg)
  Start
  Write(devID<<1)
  Write(devReg)
  result := ReReadByte(devID)
  Stop

PUB ReReadByte(devID) : ackbits
  ' use this when the pointer or command register is already set or when it autoincrements.
  Start
  Write(devID<<1 | 1)
  result := Read(NAK)

CON  ' lock methods. if needed.  All buss activity should be routed via this object for locks to be effective.
' SetLock should normally precede Start or OpenDevice, ClearLock should follow the Stop or a method that ends in Stop.
PUB SetLock                     ' Generally call this before Start, or before an extended buss transaction
  if lockID == 0                ' note, this does not report error if no locks are availabe, it just proceeds without lock
    lockID := locknew + 1
  if lockID
    repeat until not lockset(lockID-1)
  return lockID   ' zero if none, 1 to 8 if used

PUB ClearLock                   ' Generally call this after Stop, or after an extended buss transaction
  if lockID
    lockclr(lockID - 1)

PUB ReturnLock                   ' When Lock is no longer needed
  if lockID
    lockret(lockID - 1)


DAT
  ' could do all of this in two longs passed in at init.
  timeout long 8_000_000   ' 0.1 second at 80 MHz
  scl byte -1
  sda byte -1
  lockID byte 0
  instance byte 0

